unit uPlayer;

interface

uses Classes, Windows, MMSystem, SysUtils;

type
  ESndError = class(Exception);

  TSoundObject = class(TObject)
  private
  protected
    FHandle:THandle;
  public
    procedure Open;virtual;abstract;
    procedure Close;virtual;abstract;
    procedure Start;virtual;abstract;
    procedure Stop;virtual;abstract;
  published
    property Handle:THandle read FHandle;
  end;

  TWaveObject = class(TSoundObject)
  private
    procedure SetupWaveFormat;
    procedure SetBitsPerSample(const Value: word);
    procedure SetChannel(const Value: word);
    procedure SetSamplePerSec(const Value: cardinal);
  protected
    FWaveFormat:TWaveFormatEx;

    procedure WaveProc(const handle:THandle;
                      const msg:UINT;
                      const dwInstance:cardinal;
                      const dwParam1,dwParam2:cardinal);virtual;abstract;
  public
    constructor Create;virtual;
  published
    property Channel:word read FWaveFormat.nChannels write SetChannel;
    property SamplePerSec:cardinal read FWaveFormat.nSamplesPerSec write SetSamplePerSec;
    property BitsPerSample:word read FWaveFormat.wBitsPerSample write SetBitsPerSample;
  end;



  TDataRequiredEvent=procedure(sender:TObject;
                            const Buffer:pointer;
                            const BufferSize:cardinal;
                            var BytesInBuffer:cardinal) of object;

  TPlayer=class(TWaveObject)
  private
    FBuffer1,FBuffer2,FCurrentBuffer:PWaveHdr;

    FPlaying: boolean;
    FOnDataRequired: TDataRequiredEvent;
    FLeftVolume,FRightVolume:word;
    
    procedure SetPlaying(const Value: boolean);
    procedure SwapBuffers;
    procedure SetOnDataRequired(const Value: TDataRequiredEvent);
    procedure WriteData;

    function  GetCurrentPosTime: cardinal;
    function  GetCurrentPosBytes: cardinal;
    procedure SetLeftVolume(const Value: word);
    function  GetLeftVolume: word;
    procedure SetRightVolume(const Value: word);
    function  GetRightVolume: word;
  protected
    procedure DoDataRequired(const Buffer:pointer;
                           const BufferSize:cardinal;
                           var BytesInBuffer:cardinal);virtual;
    procedure WaveProc(const handle:THandle;
                      const msg:UINT;
                      const dwInstance:cardinal;
                      const dwParam1,dwParam2:cardinal);override;
  public
    constructor Create;override;
    destructor Destroy;override;
    procedure Open;override;
    procedure Close;override;
    procedure Start;override;
    procedure Stop;override;

    procedure Pause;
    procedure Resume;
  published
    property CurrentPosTime:cardinal read GetCurrentPosTime;
    property CurrentPosBytes:cardinal read GetCurrentPosBytes;

    property LeftVolume:word read GetLeftVolume write SetLeftVolume;
    property RightVolume:word read GetRightVolume write SetRightVolume;

    property Playing:boolean read FPlaying write SetPlaying;
    property OnDataRequired:TDataRequiredEvent read FOnDataRequired write SetOnDataRequired;
  end;

implementation

const MAX_BUFFER_SIZE = 4*1024;
      PLAYBACK_BUFFER_SIZE = 64*1024;

{ TPlayer }

procedure TPlayer.Close;
begin
  if FHandle<>0 then
  begin
    Stop;
    waveOutUnPrepareHeader(Handle,FBuffer1,Sizeof(TWaveHdr));
    waveOutUnPrepareHeader(Handle,FBuffer2,Sizeof(TWaveHdr));
    WaveOutClose(FHandle);
    FHandle:=0;
  end;
end;

procedure _WaveOutProc(Handle:HWAVEOUT;uMsg:UINT;
                      dwInstance:DWORD;
                      dwParam1,dwParam2:DWORD);stdcall;
begin
  TPlayer(dwInstance).WaveProc(handle,
                 uMsg,
                 dwInstance,
                 dwParam1,
                 dwParam2);
end;

constructor TPlayer.Create;
begin
  inherited;
  new(FBuffer1);
  ZeroMemory(FBuffer1,sizeOf(TWaveHdr));
  GetMem(FBuffer1.lpData,PLAYBACK_BUFFER_SIZE);
  FBuffer1.dwBufferLength:=PLAYBACK_BUFFER_SIZE;

  new(FBuffer2);
  ZeroMemory(FBuffer2,sizeOf(TWaveHdr));
  GetMem(FBuffer2.lpData,PLAYBACK_BUFFER_SIZE);
  FBuffer2.dwBufferLength:=PLAYBACK_BUFFER_SIZE;
end;

destructor TPlayer.Destroy;
begin
  Close;
  FreeMem(FBuffer1.lpData,PLAYBACK_BUFFER_SIZE);
  FreeMem(FBuffer2.lpData,PLAYBACK_BUFFER_SIZE);
  dispose(FBuffer1);
  dispose(FBuffer2);
  inherited;
end;

procedure TPlayer.DoDataRequired(const Buffer: pointer;
  const BufferSize: cardinal; var BytesInBuffer: cardinal);
begin
  if Assigned(FOnDataRequired) then
    FOnDataRequired(self,Buffer,BufferSize,BytesInBuffer);
end;

function TPlayer.GetCurrentPosBytes: cardinal;
var posInfo:TMMTime;
begin
  if (Handle<>0) then
  begin
    ZeroMemory(@posInfo,sizeof(TMMTime));
    PosInfo.wType:=TIME_BYTES;
    waveOutGetPosition(Handle,@posInfo,sizeof(TMMTime));
    result:=posInfo.cb;
  end else
    result:=0;
end;

function TPlayer.GetCurrentPosTime: cardinal;
var posInfo:TMMTime;
begin
  result:=0;
  if Handle<>0 then
  begin
    PosInfo.wType:=TIME_MS;
    waveOutGetPosition(Handle,@posInfo,sizeof(TMMTime));
    result:=posInfo.ms;
  end;
end;

function TPlayer.GetLeftVolume: word;
var dwVolume:cardinal;
begin
  waveOutGetVolume(FHandle,@dwVolume);
  FLeftVolume:=LoWord(dwVolume);
  FRightVolume:=HiWord(dwVolume);
  result:=FLeftVolume;
end;

function TPlayer.GetRightVolume: word;
var dwVolume:cardinal;
begin
  waveOutGetVolume(FHandle,@dwVolume);
  FLeftVolume:=LoWord(dwVolume);
  FRightVolume:=HiWord(dwVolume);
  result:=FRightVolume;
end;

procedure TPlayer.Open;
var ahandle:HWAVEOUT;
    status:MMResult;
    statusStr:string;
begin
  if Handle=0 then
  begin
    status:=WaveOutOpen(@aHandle,
                        WAVE_MAPPER,
                        @FWaveFormat,
                        cardinal(@_WaveOutProc),
                        cardinal(Self),
                        CALLBACK_FUNCTION);

    FHandle:=aHandle;
    if status<>MMSYSERR_NOERROR then
    begin
      setlength(statusStr,MAXERRORLENGTH);
      waveOutGetErrorText(status,pChar(statusStr),
                       MAXERRORLENGTH);
      raise ESndError.Create(statusStr);
    end;

    WaveOutPrepareHeader(Handle,FBuffer1,sizeof(TWaveHdr));
    WaveOutPrepareHeader(Handle,FBuffer2,sizeof(TWaveHdr));
  end;
end;

procedure TPlayer.Pause;
begin
  if FHandle<>0 then
    WaveOutPause(FHandle);
end;

procedure TPlayer.Resume;
begin
  if FHandle<>0 then
    WaveOutRestart(FHandle);
end;

procedure TPlayer.SetLeftVolume(const Value: word);
var dwVolume:cardinal;
begin
  FLeftVolume:=value;
  dwVolume:=(FRightVolume shl 16) or FLeftVolume;
  waveOutSetVolume(FHandle,dwVolume);
end;

procedure TPlayer.SetOnDataRequired(const Value: TDataRequiredEvent);
begin
  FOnDataRequired := Value;
end;

procedure TPlayer.SetPlaying(const Value: boolean);
begin
  Stop;
end;

procedure TPlayer.SetRightVolume(const Value: word);
var dwVolume:cardinal;
begin
  FRightVolume:=value;
  dwVolume:=(FRightVolume shl 16) or FLeftVolume;
  waveOutSetVolume(FHandle,dwVolume);
end;

procedure TPlayer.Start;
begin
  if Handle<>0 then
  begin
    Stop;
    FCurrentBuffer:=FBuffer1;
    FBuffer1.dwBufferLength:=PLAYBACK_BUFFER_SIZE;
    FBuffer2.dwBufferLength:=PLAYBACK_BUFFER_SIZE;
    WriteData;
    FPlaying:=true;
  end;
end;

procedure TPlayer.Stop;
begin
  FPlaying:=false;
  if FHandle<>0 then
     WaveOutReset(FHandle);
end;

procedure TPlayer.SwapBuffers;
begin
  if FCurrentBuffer=FBuffer1 then
    FCurrentBuffer:=FBuffer2
  else
    FCurrentBuffer:=FBuffer1;
end;

procedure TPlayer.WaveProc(const handle:THandle;
                      const msg:UINT;
                      const dwInstance:cardinal;
                      const dwParam1,dwParam2:cardinal);
begin
  case msg of
    WOM_DONE:begin
               if FPlaying then
               begin
                 SwapBuffers;
                 WriteData;
               end;
             end;
  end;
end;

procedure TPlayer.WriteData;
var ActBytesInBuffer:cardinal;
begin
  ActBytesInBuffer:=0;
  DoDataRequired(FCurrentBuffer.lpData,
                 FCurrentBuffer.dwBufferLength,
                 ActBytesInBuffer);

  if ActBytesInBuffer=0 then
  begin
    FPlaying:=false;
    exit;
  end;

  if ActBytesInBuffer<FCurrentBuffer.dwBufferLength then
  begin
    FCurrentBuffer.dwBufferLength:=ActBytesInBuffer;
    WaveOutWrite(Handle,FCurrentBuffer,sizeof(TWaveHdr));
    FPlaying:=false;
  end else
    WaveOutWrite(Handle,FCurrentBuffer,sizeof(TWaveHdr));
end;

{ TWaveObject }

constructor TWaveObject.Create;
begin
  FWaveFormat.nChannels:=1;
  FWaveFormat.nSamplesPerSec:=11025;
  FWaveFormat.wBitsPerSample:=8;
  SetupWaveFormat;
end;

procedure TWaveObject.SetBitsPerSample(const Value: word);
begin
  FWaveFormat.wBitsPerSample:=value;
  SetupWaveFormat;
end;

procedure TWaveObject.SetChannel(const Value: word);
begin
  FWaveFormat.nChannels:=value;
  SetupWaveFormat;
end;

procedure TWaveObject.SetSamplePerSec(const Value: cardinal);
begin
  FWaveFormat.nSamplesPerSec:=value;
  SetupWaveFormat;
end;

procedure TWaveObject.SetupWaveFormat;
begin
  with FWaveFormat do
  begin
    wFormatTag:=WAVE_FORMAT_PCM;
    nBlockAlign:=nChannels*wBitsPerSample div 8;
    nAvgBytesPerSec:=nSamplesPerSec*nBlockAlign;
    cbSize:=0;
  end;
end;


end.
